"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DeclarationIndex = void 0;
var tslib_1 = require("tslib");
var lodash_1 = require("lodash");
var path_1 = require("path");
var DeclarationInfo_1 = require("./declarations/DeclarationInfo");
var ModuleDeclaration_1 = require("./declarations/ModuleDeclaration");
var AllExport_1 = require("./exports/AllExport");
var AssignedExport_1 = require("./exports/AssignedExport");
var NamedExport_1 = require("./exports/NamedExport");
var File_1 = require("./resources/File");
var Module_1 = require("./resources/Module");
var Namespace_1 = require("./resources/Namespace");
var TypescriptHeroGuards_1 = require("./type-guards/TypescriptHeroGuards");
var PathHelpers_1 = require("./utilities/PathHelpers");
/**
 * Returns the name of the node folder. Is used as the library name for indexing.
 * (e.g. ./node_modules/webpack returns webpack)
 *
 * @param {string} path
 * @returns {string}
 */
function getNodeLibraryName(path) {
    var dirs = path.split(/\/|\\/);
    var nodeIndex = dirs.indexOf('node_modules');
    return PathHelpers_1.normalizeFilename(dirs.slice(nodeIndex + 1).join('/'))
        .replace(new RegExp("/(index|" + dirs[nodeIndex + 1] + "|" + dirs[dirs.length - 2] + ")$"), '');
}
/**
 * Global index of declarations. Contains declarations and origins.
 * Provides reverse index for search and declaration info for imports.
 *
 * @export
 * @class DeclarationIndex
 */
var DeclarationIndex = /** @class */ (function () {
    function DeclarationIndex(parser, rootPath) {
        this.parser = parser;
        this.rootPath = rootPath;
        this.building = false;
        /**
         * Hash of parsed resources. Contains all parsed files / namespaces / declarations
         * of the current workspace.
         *
         * @private
         * @type {Resources}
         * @memberof DeclarationIndex
         */
        this.parsedResources = Object.create(null);
    }
    Object.defineProperty(DeclarationIndex.prototype, "indexReady", {
        /**
         * Indicator if the first index was loaded and calculated or not.
         *
         * @readonly
         * @type {boolean}
         * @memberof DeclarationIndex
         */
        get: function () {
            return this._index !== undefined; // && this._indexWithBarrels !== undefined;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(DeclarationIndex.prototype, "index", {
        /**
         * Reverse index of the declarations.
         *
         * @readonly
         * @type {({ [declaration: string]: DeclarationInfo[] } | undefined)}
         * @memberof DeclarationIndex
         */
        get: function () {
            return this._index;
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(DeclarationIndex.prototype, "declarationInfos", {
        /**
         * List of all declaration information. Contains the typescript declaration and the
         * "from" information (from where the symbol is imported).
         *
         * @readonly
         * @type {DeclarationInfo[]}
         * @memberof DeclarationIndex
         */
        get: function () {
            var _this = this;
            return Object
                .keys(this.index)
                .sort()
                .reduce(function (all, key) { return all.concat(_this.index[key]); }, []);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Calculates the differences between two indices. Calculates removed, added and updated declarations.
     * The updated declarations are calculated and all declarations that the new index contains are inserted in the list.
     *
     * @static
     * @param {{ [declaration: string]: DeclarationInfo[] }} oldIndex
     * @param {{ [declaration: string]: DeclarationInfo[] }} newIndex
     * @returns {IndexDelta}
     * @memberof DeclarationIndex
     */
    DeclarationIndex.calculateIndexDelta = function (oldIndex, newIndex) {
        var oldKeys = Object.keys(oldIndex);
        var newKeys = Object.keys(newIndex);
        return {
            added: lodash_1.difference(newKeys, oldKeys).reduce(function (obj, currentKey) {
                obj[currentKey] = newIndex[currentKey];
                return obj;
            }, {}),
            updated: lodash_1.intersection(oldKeys, newKeys).reduce(function (obj, currentKey) {
                var old = oldIndex[currentKey];
                var neu = newIndex[currentKey];
                if (lodash_1.differenceWith(neu, old, lodash_1.isEqual).length > 0 || lodash_1.differenceWith(old, neu, lodash_1.isEqual).length > 0) {
                    obj[currentKey] = neu;
                }
                return obj;
            }, {}),
            deleted: lodash_1.difference(oldKeys, newKeys),
        };
    };
    /**
     * Resets the whole index. Does delete everything. Period.
     * Is useful for unit testing or similar things.
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.reset = function () {
        this.parsedResources = Object.create(null);
        this._index = undefined;
    };
    /**
     * Tells the index to build a new index.
     * Can be canceled with a cancellationToken.
     *
     * @param {string[]} filePathes
     * @returns {Promise<void>}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.buildIndex = function (filePathes) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var parsed, _a, _b, e_1;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (this.building) {
                            return [2 /*return*/];
                        }
                        _c.label = 1;
                    case 1:
                        _c.trys.push([1, 5, 6, 7]);
                        this.building = true;
                        return [4 /*yield*/, this.parser.parseFiles(filePathes, this.rootPath)];
                    case 2:
                        parsed = _c.sent();
                        _a = this;
                        return [4 /*yield*/, this.parseResources(parsed)];
                    case 3:
                        _a.parsedResources = _c.sent();
                        _b = this;
                        return [4 /*yield*/, this.createIndex(this.parsedResources)];
                    case 4:
                        _b._index = _c.sent();
                        return [3 /*break*/, 7];
                    case 5:
                        e_1 = _c.sent();
                        throw e_1;
                    case 6:
                        this.building = false;
                        return [7 /*endfinally*/];
                    case 7: return [2 /*return*/];
                }
            });
        });
    };
    /**
     * Is called when file events happen. Does reindex for the changed files and creates a new index.
     * Returns the differences for the new index.
     *
     * @param {FileEvent[]} changes
     * @returns {Promise<IndexDelta>}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.reindexForChanges = function (changes) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var rebuildResources, removeResources, rebuildFiles, _i, _a, change, filePath, resource, _b, _c, file, _d, _e, change, filePath, resource, _f, _g, file, resources, _h, _j, removeResources_1, del, _k, _l, key, old, _m;
            return tslib_1.__generator(this, function (_o) {
                switch (_o.label) {
                    case 0:
                        rebuildResources = [];
                        removeResources = [];
                        rebuildFiles = [];
                        for (_i = 0, _a = changes.deleted; _i < _a.length; _i++) {
                            change = _a[_i];
                            filePath = PathHelpers_1.normalizePathUri(change);
                            resource = '/' + PathHelpers_1.normalizeFilename(path_1.relative(this.rootPath, filePath));
                            if (removeResources.indexOf(resource) < 0) {
                                removeResources.push(resource);
                            }
                            for (_b = 0, _c = this.getExportedResources(resource); _b < _c.length; _b++) {
                                file = _c[_b];
                                if (rebuildFiles.indexOf(file) < 0) {
                                    rebuildFiles.push(file);
                                }
                            }
                        }
                        for (_d = 0, _e = (changes.created || []).concat(changes.updated); _d < _e.length; _d++) {
                            change = _e[_d];
                            filePath = PathHelpers_1.normalizePathUri(change);
                            resource = '/' + PathHelpers_1.normalizeFilename(path_1.relative(this.rootPath, filePath));
                            if (rebuildResources.indexOf(resource) < 0) {
                                rebuildResources.push(resource);
                            }
                            if (rebuildFiles.indexOf(filePath) < 0) {
                                rebuildFiles.push(filePath);
                            }
                            for (_f = 0, _g = this.getExportedResources(resource); _f < _g.length; _f++) {
                                file = _g[_f];
                                if (rebuildFiles.indexOf(file) < 0) {
                                    rebuildFiles.push(file);
                                }
                            }
                        }
                        _h = this.parseResources;
                        return [4 /*yield*/, this.parser.parseFiles(rebuildFiles, this.rootPath)];
                    case 1: return [4 /*yield*/, _h.apply(this, [_o.sent()])];
                    case 2:
                        resources = _o.sent();
                        for (_j = 0, removeResources_1 = removeResources; _j < removeResources_1.length; _j++) {
                            del = removeResources_1[_j];
                            delete this.parsedResources[del];
                        }
                        for (_k = 0, _l = Object.keys(resources); _k < _l.length; _k++) {
                            key = _l[_k];
                            this.parsedResources[key] = resources[key];
                        }
                        old = this._index || {};
                        _m = this;
                        return [4 /*yield*/, this.createIndex(this.parsedResources)];
                    case 3:
                        _m._index = _o.sent();
                        return [2 /*return*/, DeclarationIndex.calculateIndexDelta(old, this._index)];
                }
            });
        });
    };
    /**
     * Returns a list of files that export a certain resource (declaration).
     *
     * @private
     * @param {string} resourceToCheck
     * @returns {string[]}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.getExportedResources = function (resourceToCheck) {
        var _this = this;
        var resources = [];
        Object
            .keys(this.parsedResources)
            .forEach(function (key) {
            var resource = _this.parsedResources[key];
            if (resource instanceof File_1.File && _this.doesExportResource(resource, resourceToCheck)) {
                resources.push(resource.filePath);
            }
        });
        return resources;
    };
    /**
     * Checks if a file does export another resource.
     * (i.e. export ... from ...)
     *
     * @private
     * @param {File} resource The file that is checked
     * @param {string} resourcePath The resource that is searched for
     * @returns {boolean}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.doesExportResource = function (resource, resourcePath) {
        var exportsResource = false;
        for (var _i = 0, _a = resource.exports; _i < _a.length; _i++) {
            var ex = _a[_i];
            if (exportsResource) {
                break;
            }
            if (ex instanceof AllExport_1.AllExport || ex instanceof NamedExport_1.NamedExport) {
                var exported = '/' + PathHelpers_1.toPosix(path_1.relative(this.rootPath, path_1.normalize(path_1.join(resource.parsedPath.dir, ex.from))));
                exportsResource = exported === resourcePath;
            }
        }
        return exportsResource;
    };
    /**
     * Does parse the resources (symbols and declarations) of a given file.
     * Can be cancelled with the token.
     *
     * @private
     * @param {File[]} [files=[]]
     * @returns {Promise<Resources>}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.parseResources = function (files) {
        if (files === void 0) { files = []; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var parsedResources, _i, files_1, file, _a, _b, resource, libname, _c, _d, key, resource;
            return tslib_1.__generator(this, function (_e) {
                parsedResources = Object.create(null);
                for (_i = 0, files_1 = files; _i < files_1.length; _i++) {
                    file = files_1[_i];
                    if (file.filePath.indexOf('typings') > -1 || file.filePath.indexOf('node_modules/@types') > -1) {
                        for (_a = 0, _b = file.resources; _a < _b.length; _a++) {
                            resource = _b[_a];
                            parsedResources[resource.identifier] = resource;
                        }
                    }
                    else if (file.filePath.indexOf('node_modules') > -1) {
                        libname = getNodeLibraryName(file.filePath);
                        parsedResources[libname] = file;
                    }
                    else {
                        parsedResources[file.identifier] = file;
                    }
                }
                for (_c = 0, _d = Object.keys(parsedResources).sort(function (k1, k2) { return k2.length - k1.length; }); _c < _d.length; _c++) {
                    key = _d[_c];
                    resource = parsedResources[key];
                    resource.declarations = resource.declarations.filter(function (o) { return TypescriptHeroGuards_1.isExportableDeclaration(o) && o.isExported; });
                    this.processResourceExports(parsedResources, resource);
                }
                return [2 /*return*/, parsedResources];
            });
        });
    };
    /**
     * Creates a reverse index out of the give resources.
     * Can be cancelled with the token.
     *
     * @private
     * @param {Resources} resources
     * @returns {Promise<ResourceIndex>}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.createIndex = function (resources) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var index, _i, _a, key, resource, _loop_1, _b, _c, declaration;
            return tslib_1.__generator(this, function (_d) {
                index = Object.create(null);
                for (_i = 0, _a = Object.keys(resources); _i < _a.length; _i++) {
                    key = _a[_i];
                    resource = resources[key];
                    if (resource instanceof Namespace_1.Namespace || resource instanceof Module_1.Module) {
                        if (!index[resource.name]) {
                            index[resource.name] = [];
                        }
                        index[resource.name].push(new DeclarationInfo_1.DeclarationInfo(new ModuleDeclaration_1.ModuleDeclaration(resource.getNamespaceAlias(), resource.start, resource.end), resource.name));
                    }
                    _loop_1 = function (declaration) {
                        if (!index[declaration.name]) {
                            index[declaration.name] = [];
                        }
                        var from = key.replace(/\/index$/, '') || '/';
                        if (!index[declaration.name].some(function (o) { return o.declaration.constructor === declaration.constructor && o.from === from; })) {
                            index[declaration.name].push(new DeclarationInfo_1.DeclarationInfo(declaration, from));
                        }
                    };
                    for (_b = 0, _c = resource.declarations; _b < _c.length; _b++) {
                        declaration = _c[_b];
                        _loop_1(declaration);
                    }
                }
                return [2 /*return*/, index];
            });
        });
    };
    /**
     * Process all exports of a the parsed resources. Does move the declarations accordingly to their
     * export nature.
     *
     * @private
     * @param {Resources} parsedResources
     * @param {Resource} resource
     * @param {Resource[]} [processedResources=[]]
     * @returns {void}
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.processResourceExports = function (parsedResources, resource, processedResources) {
        if (processedResources === void 0) { processedResources = []; }
        if (processedResources.indexOf(resource) >= 0 || resource.exports.length === 0) {
            return;
        }
        processedResources.push(resource);
        for (var _i = 0, _a = resource.exports; _i < _a.length; _i++) {
            var ex = _a[_i];
            if (resource instanceof File_1.File && (ex instanceof NamedExport_1.NamedExport || ex instanceof AllExport_1.AllExport)) {
                if (!ex.from) {
                    return;
                }
                var sourceLib = path_1.resolve(resource.parsedPath.dir, ex.from);
                if (sourceLib.indexOf('node_modules') > -1) {
                    sourceLib = getNodeLibraryName(sourceLib);
                }
                else {
                    sourceLib = '/' + PathHelpers_1.normalizeFilename(path_1.relative(this.rootPath, sourceLib));
                }
                if (!parsedResources[sourceLib]) {
                    return;
                }
                var exportedLib = parsedResources[sourceLib];
                this.processResourceExports(parsedResources, exportedLib, processedResources);
                if (ex instanceof AllExport_1.AllExport) {
                    this.processAllFromExport(resource, exportedLib);
                }
                else {
                    this.processNamedFromExport(ex, resource, exportedLib);
                }
            }
            else {
                if (ex instanceof AssignedExport_1.AssignedExport) {
                    for (var _b = 0, _c = ex.exported.filter(function (o) { return !TypescriptHeroGuards_1.isExportableDeclaration(o); }); _b < _c.length; _b++) {
                        var lib = _c[_b];
                        this.processResourceExports(parsedResources, lib, processedResources);
                    }
                    this.processAssignedExport(ex, resource);
                }
                else if (ex instanceof NamedExport_1.NamedExport && ex.from && parsedResources[ex.from]) {
                    this.processResourceExports(parsedResources, parsedResources[ex.from], processedResources);
                    this.processNamedFromExport(ex, resource, parsedResources[ex.from]);
                }
            }
        }
    };
    /**
     * Processes an all export, does move the declarations accordingly.
     * (i.e. export * from './myFile')
     *
     * @private
     * @param {Resource} exportingLib
     * @param {Resource} exportedLib
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.processAllFromExport = function (exportingLib, exportedLib) {
        var _a;
        (_a = exportingLib.declarations).push.apply(_a, exportedLib.declarations);
        exportedLib.declarations = [];
    };
    /**
     * Processes a named export, does move the declarations accordingly.
     * (i.e. export {MyClass} from './myFile')
     *
     * @private
     * @param {NamedExport} tsExport
     * @param {Resource} exportingLib
     * @param {Resource} exportedLib
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.processNamedFromExport = function (tsExport, exportingLib, exportedLib) {
        exportedLib.declarations
            .forEach(function (o) {
            if (!tsExport.specifiers) {
                return;
            }
            var ex = tsExport.specifiers.find(function (s) { return s.specifier === o.name; });
            if (!ex) {
                return;
            }
            exportedLib.declarations.splice(exportedLib.declarations.indexOf(o), 1);
            if (ex.alias) {
                o.name = ex.alias;
            }
            exportingLib.declarations.push(o);
        });
    };
    /**
     * Processes an assigned export, does move the declarations accordingly.
     * (i.e. export = namespaceName)
     *
     * @private
     * @param {AssignedExport} tsExport
     * @param {Resource} exportingLib
     *
     * @memberof DeclarationIndex
     */
    DeclarationIndex.prototype.processAssignedExport = function (tsExport, exportingLib) {
        tsExport.exported.forEach(function (exported) {
            var _a;
            if (TypescriptHeroGuards_1.isExportableDeclaration(exported)) {
                exportingLib.declarations.push(exported);
            }
            else {
                (_a = exportingLib.declarations).push.apply(_a, exported.declarations.filter(function (o) { return TypescriptHeroGuards_1.isExportableDeclaration(o) && o.isExported; }));
                exported.declarations = [];
            }
        });
    };
    return DeclarationIndex;
}());
exports.DeclarationIndex = DeclarationIndex;
