"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TypescriptParser = void 0;
var tslib_1 = require("tslib");
var fs_1 = require("fs");
var path_1 = require("path");
var typescript_1 = require("typescript");
var class_parser_1 = require("./node-parser/class-parser");
var enum_parser_1 = require("./node-parser/enum-parser");
var export_parser_1 = require("./node-parser/export-parser");
var function_parser_1 = require("./node-parser/function-parser");
var identifier_parser_1 = require("./node-parser/identifier-parser");
var import_parser_1 = require("./node-parser/import-parser");
var interface_parser_1 = require("./node-parser/interface-parser");
var module_parser_1 = require("./node-parser/module-parser");
var traverse_ast_1 = require("./node-parser/traverse-ast");
var type_alias_parser_1 = require("./node-parser/type-alias-parser");
var variable_parser_1 = require("./node-parser/variable-parser");
var File_1 = require("./resources/File");
/**
 * Magic.happens('here');
 * This class is the parser of the whole extension. It uses the typescript compiler to parse a file or given
 * source code into the token stream and therefore into the AST of the source. Afterwards an array of
 * resources is generated and returned.
 *
 * @export
 * @class TypescriptParser
 */
var TypescriptParser = /** @class */ (function () {
    function TypescriptParser() {
    }
    /**
     * Parses the given source into an anonymous File resource.
     * Mainly used to parse source code of a document.
     *
     * @param {string} source
     * @param {ScriptKind} [scriptKind=ScriptKind.TS]
     * @returns {Promise<File>}
     *
     * @memberof TsResourceParser
     */
    TypescriptParser.prototype.parseSource = function (source, scriptKind) {
        if (scriptKind === void 0) { scriptKind = typescript_1.ScriptKind.TS; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.parseTypescript(typescript_1.createSourceFile('inline.tsx', source, typescript_1.ScriptTarget.ES2015, true, scriptKind), '/')];
                    case 1: return [2 /*return*/, _a.sent()];
                }
            });
        });
    };
    /**
     * Parses a single file into a parsed file.
     *
     * @param {string} filePath
     * @param {string} rootPath
     * @returns {Promise<File>}
     *
     * @memberof TsResourceParser
     */
    TypescriptParser.prototype.parseFile = function (filePath, rootPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var parse;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.parseFiles([filePath], rootPath)];
                    case 1:
                        parse = _a.sent();
                        return [2 /*return*/, parse[0]];
                }
            });
        });
    };
    /**
     * Parses multiple files into parsed files.
     *
     * @param {string[]} filePathes
     * @param {string} rootPath
     * @returns {Promise<File[]>}
     *
     * @memberof TsResourceParser
     */
    TypescriptParser.prototype.parseFiles = function (filePathes, rootPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _this = this;
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, filePathes
                        .map(function (o) {
                        var scriptKind = typescript_1.ScriptKind.Unknown;
                        var parsed = path_1.parse(o);
                        switch (parsed.ext.toLowerCase()) {
                            case 'js':
                                scriptKind = typescript_1.ScriptKind.JS;
                                break;
                            case 'jsx':
                                scriptKind = typescript_1.ScriptKind.JSX;
                                break;
                            case 'ts':
                                scriptKind = typescript_1.ScriptKind.TS;
                                break;
                            case 'tsx':
                                scriptKind = typescript_1.ScriptKind.TSX;
                                break;
                        }
                        return typescript_1.createSourceFile(o, fs_1.readFileSync(o).toString(), typescript_1.ScriptTarget.ES2015, true, scriptKind);
                    })
                        .map(function (o) { return _this.parseTypescript(o, rootPath); })];
            });
        });
    };
    /**
     * Parses the typescript source into the file instance. Calls .parse afterwards to
     * get the declarations and other information about the source.
     *
     * @private
     * @param {SourceFile} source
     * @param {string} rootPath
     * @returns {TsFile}
     *
     * @memberof TsResourceParser
     */
    TypescriptParser.prototype.parseTypescript = function (source, rootPath) {
        var file = new File_1.File(source.fileName, rootPath, source.getStart(), source.getEnd());
        var syntaxList = source.getChildAt(0);
        this.parse(file, syntaxList);
        return file;
    };
    /**
     * Recursive function that runs through the AST of a source and parses the nodes.
     * Creates the class / function / etc declarations and instanciates a new module / namespace
     * resource if needed.
     *
     * @private
     * @param {Resource} resource
     * @param {Node} node
     *
     * @memberof TsResourceParser
     */
    TypescriptParser.prototype.parse = function (resource, root) {
        var modules = [{ moduleRoot: root, moduleResource: resource }];
        var _loop_1 = function (iter) {
            var moduleRoot = iter.moduleRoot, moduleResource = iter.moduleResource;
            traverse_ast_1.traverseAst(moduleRoot, function (node) {
                switch (node.kind) {
                    case typescript_1.SyntaxKind.ImportDeclaration:
                    case typescript_1.SyntaxKind.ImportEqualsDeclaration:
                        import_parser_1.parseImport(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.ExportDeclaration:
                    case typescript_1.SyntaxKind.ExportAssignment:
                        export_parser_1.parseExport(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.EnumDeclaration:
                        enum_parser_1.parseEnum(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.TypeAliasDeclaration:
                        type_alias_parser_1.parseTypeAlias(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.FunctionDeclaration:
                        function_parser_1.parseFunction(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.VariableStatement:
                        variable_parser_1.parseVariable(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.InterfaceDeclaration:
                        interface_parser_1.parseInterface(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.ClassDeclaration:
                        class_parser_1.parseClass(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.Identifier:
                        identifier_parser_1.parseIdentifier(moduleResource, node);
                        break;
                    case typescript_1.SyntaxKind.ModuleDeclaration:
                        modules.push({
                            moduleRoot: node,
                            moduleResource: module_parser_1.parseModule(moduleResource, node),
                        });
                        break;
                    default:
                        break;
                }
            }, function (node) {
                switch (node.kind) {
                    case typescript_1.SyntaxKind.ClassDeclaration:
                    case typescript_1.SyntaxKind.ModuleDeclaration:
                    case typescript_1.SyntaxKind.FunctionDeclaration:
                        return true;
                    default:
                        return false;
                }
            });
        };
        for (var iter = modules.shift(); iter !== undefined; iter = modules.shift()) {
            _loop_1(iter);
        }
    };
    return TypescriptParser;
}());
exports.TypescriptParser = TypescriptParser;
